module control
#(
    parameter WORD = 1,
    parameter HOR = 28,

    parameter ADDR_W = $clog2(HOR),
    parameter Q = 3     // number of fractional bits
)(
//--------------Software interface  ---------------
    input logic clock,
    input logic reset,
    
    //----------- image data ----------------------
    input logic image_load_done,
    input logic image_data,
    input logic image_write_en,
    input logic [9:0] img_wr_addr,

    input logic [3:0] dense_address,
    output logic signed [31:0] dense_output,
    output logic dense_done,

    // ------------ conv ->  Pool handshake debug
    output logic conv_valid_out, // valid signal to pool layer
    output logic signed [31:0] mac_out, // convolution output to the pool layer
    output logic pool_valid_out,
    output logic signed [7:0] pooled_out,
    output logic [16:0] counter_dbg

    // conv layer debug
   /* output logic [WORD-1:0] mem0_out [0:HOR-1],
    output logic [WORD-1:0] mem1_out [0:HOR-1],
    output logic [WORD-1:0] mem2_out [0:HOR-1],
    output logic [WORD-1:0] mem3_out [0:HOR-1],
    output logic [WORD-1:0] top_line_out [0:2],
    output logic [WORD-1:0] mid_line_out [0:2],
    output logic [WORD-1:0] bot_line_out [0:2],*/
    // pool layer debug
   // output logic signed [7:0] top_line_pool [0:1],
   // output logic signed [7:0] mid_line_pool [0:1],
   // output logic [4:0] hcount, 
);

// ----------------- Internal signal ---------------------
logic [6:0] wgt_offset;
logic rd_img_en; // enable signal to convolution module
logic rd_wgt_en; // enable signal to convolution module
logic [4:0] hcount_in;
logic [9:0] counter;
logic count_trig;
logic conv_valid;
logic dense_en;
logic [3:0] times_out;
logic [14:0] dense_wgt_addr;
logic signed [7:0] dense_weight;
logic [6:0] rd_wgt_addr;
logic signed [7:0] rd_wgt_data;
logic signed [7:0] pooled;
logic signed [7:0] dense_in;
logic dense_valid;
logic signed [31:0] dense_data;
logic signed [7:0] bias;
logic signed [7:0] dense_bias;

assign pooled_out = pooled;

typedef enum logic [2:0] { S_LOAD, S_COMPUTE, S_OFFSET, S_PAUSE, S_DENSE, S_DONE} state_t;
state_t current_state, next_state;

assign conv_valid_out = conv_valid;
assign counter_dbg =counter;

// ----------------- Write logic for image data from software into hardware --------
logic [9:0] rd_img_addr;
logic rd_img_data;

//----------------------------- Pooling layer h_count ------------------------------
logic [4:0] pool_h_count;
always_ff @(posedge clock) begin
    if (reset) begin
        pool_h_count <= 0;
    end
    else if (conv_valid) begin
        pool_h_count <= (pool_h_count + 1) % 26;
    end
end

//-------------------------- Pool valid output-------------------- 

logic pool_valid;
assign pool_valid_out = pool_valid;
logic signed [15:0] delta;
logic in_all;
logic [4:0] phase; 
logic [5:0] row_offset;
logic [3:0] row_index;

// ----------------Pooling address to Pool_out_mem.sv -----------------------------------
logic [11:0] pool_mem_addr;
always_ff @(posedge clock) begin
    if(reset) begin
        pool_mem_addr <= 0;
    end
    else if (pool_valid) begin
        pool_mem_addr <= pool_mem_addr +1;
    end
end

///--------------- Start trigger for FSM ot move out of IDLE state -----------------------
always_ff @(posedge clock) begin
    if (reset) begin
        count_trig <= 0;
    end
    else begin
        if (image_load_done) begin
            count_trig <= 1;
        end
    end
end

always_ff @( posedge clock ) begin 
    if (reset || (next_state == S_OFFSET && current_state == S_COMPUTE)) begin // on transition from S_COMPUTE to S_OFFSET
        counter <= 0;
    end
    else if (count_trig) begin
        counter <= counter+1;
    end
end

//--------------------- Weight memory offset ------------------------------
always_ff @(posedge clock) begin
    if(reset) begin
        wgt_offset <= 0;
    end
    else if (current_state == S_OFFSET && next_state == S_COMPUTE) begin //increment on transition from S_COMPUTE TO S_OFFSET
        wgt_offset <= wgt_offset + 9;
    end
end

logic [3:0] bias_addr;
always_ff @(posedge clock) begin
    if(reset) begin
        bias_addr <= '0;
    end
    else if (current_state == S_OFFSET && next_state == S_COMPUTE) begin //increment on transition from S_COMPUTE TO S_OFFSET
        bias_addr <= bias_addr + 1;
    end
end

///--------------- Dense Weight logic ---------------------------------------
always_ff @(posedge clock) begin
    if(reset) begin
        dense_wgt_addr <= '0;
    end
    else if (next_state == S_DENSE || current_state == S_DENSE) begin
        dense_wgt_addr <= dense_wgt_addr + 1;
    end
end

// ----------------- Reading from pooling Mem logic --------------------------
logic [11:0] pool_rd;

always_ff @(posedge clock) begin
    if(reset) begin
        pool_rd <= '0;
    end
    else if (next_state == S_DENSE || current_state == S_DENSE) begin
        if (pool_rd <2027) begin
            pool_rd <= pool_rd + 1;
        end
        else begin
            pool_rd <= 0;
        end
    end
end

//--------------- Dense Writing to RAM --------------------------------------
logic [4:0] dense_wr;

always_ff @(posedge clock) begin
    if(reset) begin
        dense_wr <= 0;
    end
    else if (dense_valid) begin
        dense_wr <= dense_wr + 1;
    end
end

//---------------- Dense Bias Memory control -----------------------------

logic [3:0] dense_b_addr;

always_ff @(posedge clock) begin
    if(reset) begin
        dense_b_addr <= 0;
    end
    else if (dense_valid) begin
        dense_b_addr <= dense_b_addr + 1;
    end
end
/*integer
assign fd = $fopen(" image pixel with zero point", "w");
always_ff @(posedge clock) begin
    if(data_valid_in) begin
	$fdisplay(fd, pixel);
    end
end*/

/// ---------------------------- Control FSM -------------------------------------

always_comb begin
    // default values
    next_state = current_state;
    rd_img_en = 1'b0;
    rd_img_addr = 10'b0;
    rd_wgt_en = 1'b0;
    rd_wgt_addr = 7'b0;
    
    hcount_in = 5'b0;
    conv_valid = 1'b0;
    pool_valid = 1'b0;
    dense_en = 1'b0;
    dense_done = 1'b0;

    delta = 16'sd0;
    in_all = 1'b0;
    phase = 5'b0;
    row_offset = 6'b0;
    row_index = 4'b0;

    case (current_state)
        S_LOAD: begin
            if(image_load_done) begin
                next_state = S_COMPUTE;
            end
        end


        S_COMPUTE: begin 
            rd_img_en = ((0 < counter) && (counter < 785)) ? 1 : 0;
            rd_img_addr = (counter < 784) ? counter : 0;
            rd_wgt_en = ((0 < counter) && (counter < 10)) ? 1 : 0;
            rd_wgt_addr = (counter < 9) ? counter + wgt_offset : 9+wgt_offset; 
            hcount_in = (counter > 0) ? (counter-1) % 28 : 0;
            // logic for convolution valid signal
                delta = counter - 94;
                in_all = (counter >= 94) && (counter < 822);
                phase = (counter >= 94) ? (delta % 28) : 0;
            conv_valid = in_all && (phase < 26);
            // logic for pooling valid signal
                row_offset = (counter >= 183) ? ((counter - 183) % 56): 6'b0;
                row_index  = (counter >= 183) ? ((counter - 183) / 56): 4'b0;
            pool_valid = (counter >= 183) ? ( (row_index < 13)&& (row_offset < 26) && ((row_offset % 2) == 0)): 1'b0;

            next_state = (counter == 879) ? S_OFFSET : S_COMPUTE;
                
        end

        S_OFFSET : begin
            pool_valid = 0;
            next_state = (wgt_offset == 99) ? S_PAUSE : S_COMPUTE;
        end

        S_PAUSE : begin
                next_state = S_DENSE;
        end

        S_DENSE : begin
            dense_en = 1;
            next_state = (times_out < 10) ? S_DENSE : S_DONE;
        end

        S_DONE: begin
            dense_done = 1;
            next_state = S_DONE; //^^CAN ALSO BE S_LOAD, S_DONE ASSUMING ONE IMAGE?
        end

        default: begin next_state= S_LOAD;
        dense_done = 0;
        dense_en = 0;
        rd_img_en = 0;
        rd_img_addr = 0;
        rd_wgt_en = 0;
        rd_wgt_addr = 0;
        hcount_in = 0;
        pool_valid= 0;
        conv_valid =0;
        end
    endcase
end



//--------------------- FSM update state logic --------------------------------
always_ff @(posedge clock) begin
    if (reset) begin
        current_state <= S_LOAD;
    end
    else begin
        current_state <= next_state;
    end
end

/*
always_comb begin
    $display("Weight Addr: %0d, Weight_en: %0d, weight_in: %0d", rd_wgt_addr, rd_wgt_en, rd_wgt_data);
end
*/

// ------------------ Convolution Layer instatiation logic -----------------------------
convolver conv(
    // -------------------- INPUTS -------------------------------------------------------
    .clock(clock),
    .reset(reset),
    .data_valid_in(rd_img_en),
    .pixel_data_in(rd_img_data),
    .hcount_in(hcount_in),
    .weight_en(rd_wgt_en),
    .weight_in(rd_wgt_data),
    .mac_out(mac_out),
    .pool_valid(pool_valid),
    .bias(bias)

    // ------------------- OUTPUTS --------------------------------------------------------
    /*.mem0_out(mem0_out),
    .mem1_out(mem1_out),
    .mem2_out(mem2_out),
    .mem3_out(mem3_out),
    .top_line_out(top_line_out),
    .mid_line_out(mid_line_out),
    .bot_line_out(bot_line_out),
    .hcount(hcount)*/
);

// ------------------------- Pooling Layer  instation logic ----------------------------------
pooling pool(
    // ----------- INPUTS --------------------------------------------------------------------
    .clock(clock),
    .reset(reset),
    .data_valid_in(conv_valid),
    .data_in(mac_out),
    .hcount_in(pool_h_count),

    // ----------  OUTPUTS --------------------------------------------------------------------
    .pooled_out(pooled)
    //.top_line_out(top_line_pool),
   // .mid_line_out(mid_line_pool)  
);

// ------------------------ Dense Layer instantiation logic ----------------------

/*
always_comb begin
    $display( dense_done, dense_address, dense_output);
end
*/

dense dense1(
    // ------------ INPUTS -------------------------------------------------------
    .clock(clock),
    .reset(reset),
    .data_in(dense_in),
    .weight(dense_weight),
    .enable(dense_en),
    .dense_bias(dense_bias),

    // -------------- OUTPUTS -----------------------------------------------------
    .output_data(dense_data),
    .dense_valid(dense_valid),
    .count_dbg(),
    .times_out(times_out)
);

// ----------------------- Image  Pixel Mem instantiation logic ----------------------------
image_mem image(
    .clk(clock),
    .we_b(image_write_en),
    .data_in_b(image_data),
    .addr_b(img_wr_addr),

    .addr_a(rd_img_addr),
    .data_out_a(rd_img_data)
);

// --------------------- Dense Weight Mem instatiaition logic -----------------------------------
dense_mem dense(
  .clk(clock),
  .addr_a(dense_wgt_addr),
  .data_out_a(dense_weight) 
);

// ---------------------- Conv Weight Mem instatiation logic -----------------------------------
weight_mem weights(
    .clk(clock),
    .addr_a(rd_wgt_addr),
    .data_out_a(rd_wgt_data)
);

// ----------------------  Pool out Mem instantiation logic -----------------------------------
pool_out_mem pool_out(
    .clk(clock),
    .we_a(pool_valid),
    .addr_a(pool_mem_addr),
    .data_in_a(pooled),

    .addr_b(pool_rd),
    .data_out_b(dense_in)
);

// -------------------- Dense out Mem instantiation logic ---------------------------------------
dense_out_mem dense_out(
    .clk(clock),
    .we_a(dense_valid),
    .addr_a(dense_wr),
    .data_in_a(dense_data),

    .addr_b(dense_address),
    .data_out_b(dense_output)
);

// --------------------- Conv Bias mem instatiation logic ------------------------------------------------
Cbias_mem cbias_mem1(
    .clk(clock),
    .addr_a(bias_addr),
    .data_out_a(bias)
);

Dbias_mem dbias_mem1(
    .clk(clock),
    .addr_a(dense_b_addr),
    .data_out_a(dense_bias)
);


endmodule