import tensorflow as tf
import numpy as np
import pathlib

out_dir = pathlib.Path("quartus_weights_mif")
out_dir.mkdir(exist_ok=True)
print("Exporting weights to Quartus .mif files in", out_dir)

interpreter = tf.lite.Interpreter(model_path="mnist_int8.tflite")

#print(interpreter._get_ops_details())
#
## Get number of subgraphs
#num_subgraphs = interpreter._get_ops_details()
#
## Allocate tensors for each subgraph
#for i in range(len(num_subgraphs)):
#    interpreter._interpreter.SetSubgraph(i)
interpreter.allocate_tensors()

# Get all tensor details
tensor_details = interpreter.get_tensor_details()

 # Filter for *constant* weight tensors only — ignore activation/output tensors.
 # Converter‑generated constant weights typically carry the prefix "tfl.pseudo_qconst".
weight_tensors = [
	t for t in tensor_details
	if t['quantization_parameters']['scales'].size > 0
	and t['name'].startswith("tfl.pseudo_qconst")
]

# ---- Order weight tensors by layer execution sequence --------------
# tf.lite.Interpreter._get_ops_details() yields the ops in the exact
# order the runtime will execute them.  Each op lists its input‑tensor
# indices.  Walk through that list and note the first time each
# weight‑tensor index appears; sort weights by that rank.
ops_details = interpreter._get_ops_details()  # private but fine for tooling
exec_order = {}
for rank, op in enumerate(ops_details):
    for t_idx in op["inputs"]:
        if t_idx not in exec_order:
            exec_order[t_idx] = rank  # first touch = execution order

# Give any tensors that never appear in an op input a large rank
# so they fall to the end but are still processed.
weight_tensors = sorted(
    weight_tensors,
    key=lambda t: exec_order.get(t["index"], 1_000_000)
)
# --------------------------------------------------------------------

txt = out_dir / f"weight_tensors.txt"
with open(txt, "w") as f:
    f.write(f"{weight_tensors}\n")

for idx, tensor in enumerate(weight_tensors):
    # Get the quantized weights
    weights = interpreter.get_tensor(tensor['index'])
    scale = tensor['quantization_parameters']['scales'][0]
    zero_point = tensor['quantization_parameters']['zero_points'][0]

    # Reshape conv kernels if needed (kh,kw,in_ch,out_ch)->(out_ch,in_ch,kh,kw)
    if len(weights.shape) == 4:
        weights = np.transpose(weights, (3,2,0,1))

    # Flatten the weights
    q = weights.flatten()
    depth = q.size
    width = 8

    # Encode the original tensor name in the filename for clarity
    safe_name = tensor["name"].replace("/", "_").replace(";", "_")
    mif = out_dir / f"{idx:02d}_{safe_name}.mif"
    with open(mif, "w") as f:
        f.write(f"DEPTH = {depth};\n")
        f.write(f"WIDTH = {width};\n")
        f.write("ADDRESS_RADIX = HEX;\n")
        f.write("DATA_RADIX = HEX;\n")
        f.write("CONTENT BEGIN\n")
        for addr, val in enumerate(q):
            f.write(f"    {addr:04X} : {(val & 0xFF):02X};\n")
        f.write("END;\n")
    print(f"Written {mif} (scale={scale:.6f}, zero_point={zero_point})")