'''
Utility script to:
1. Train a simple MNIST CNN
2. Preview trained model weights in human-readable form
3. Quantize to 8-bit INT8 TFLite and save
4. Evaluate quantized model accuracy using proper int8 quantization parameters
5. Export quantized weights to Quartus-compatible .mif files for FPGA initialization
'''

import numpy as np
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers
import pathlib

# --- 1) Data loading & preprocessing ---
num_classes = 10
input_shape = (28, 28, 1)
(x_train, y_train), (x_test, y_test) = keras.datasets.mnist.load_data()
# Binarize to 0/1 and cast to float32 for training
#x_train = np.where(x_train > 127, 1, 0).astype(np.float32)
#x_test  = np.where(x_test  > 127, 1, 0).astype(np.float32)
#x_train = x_train.astype(np.int8)
x_train = np.clip(x_train, -128, 127).astype(np.int8)
x_test =  np.clip(x_train, -128, 127).astype(np.int8)
# add channel dim
x_train = np.expand_dims(x_train, -1)
x_test  = np.expand_dims(x_test, -1)
# one-hot labels
y_train = keras.utils.to_categorical(y_train, num_classes)
y_test  = keras.utils.to_categorical(y_test, num_classes)
print(f"x_train: {x_train.shape}, x_test: {x_test.shape}\n")

# --- 2) Model definition & training ---
model = keras.Sequential([
    keras.Input(shape=input_shape),
    layers.Conv2D(12, (3,3), activation="relu"),
    layers.MaxPooling2D((2,2)),
    #layers.Conv2D(12, (3,3), activation="relu"),
    #layers.MaxPooling2D((2,2)),
    layers.Flatten(),
    layers.Dense(num_classes, activation="softmax"),
])
model.compile(
    loss="categorical_crossentropy",
    optimizer="adam",
    metrics=["accuracy"]
)
model.summary()

model.fit(x_train, y_train, epochs=15, batch_size=128, validation_split=0.1)

# ensure layers are built so we can access weights
#model.build(input_shape=(None, *input_shape))
## --- Fold input zero-point shift into first Conv2D bias for int8 I/O ---
#old_zp = -128
## get the first Conv2D layer
#conv_layer = model.layers[0]
#W, b = conv_layer.get_weights()
## sum the kernel weights over spatial and input-channel axes for each output filter
#kernel_sum = np.sum(W, axis=(0,1,2))
## compute bias correction: shift = -old_zp
#delta_b = -old_zp * kernel_sum
## apply folded bias
#b_folded = b + delta_b
#conv_layer.set_weights([W, b_folded])

# float model accuracy
_, float_acc = model.evaluate(x_test, y_test, verbose=0)
print(f"Float32 model test accuracy: {float_acc:.4%}\n")


# --- 4) TFLite INT8 quantization & save ---
def representative_data_gen():
    for val in tf.data.Dataset.from_tensor_slices(x_train[:100]).batch(1):
        yield [tf.cast(val, tf.float32)]

converter = tf.lite.TFLiteConverter.from_keras_model(model)
converter.optimizations = [tf.lite.Optimize.DEFAULT]
converter.representative_dataset = representative_data_gen
converter.target_spec.supported_ops = [tf.lite.OpsSet.TFLITE_BUILTINS_INT8]
converter.inference_input_type = tf.int8
converter.inference_output_type = tf.int8
# perform conversion
tflite_model = converter.convert()
with open("mnist_int8_full.tflite", "wb") as f:
    f.write(tflite_model)
print("Saved INT8 TFLite model to mnist_int8.tflite\n")

interp = tf.lite.Interpreter(model_content=tflite_model,
                             experimental_preserve_all_tensors=True)
interp.allocate_tensors()
inp = interp.get_input_details()[0]
print("New input stats:", inp["quantization"])

#--- 5) Evaluate quantized TFLite model correctly ---
def evaluate_tflite_int8(model_path, x, y):
    interpreter = tf.lite.Interpreter(model_path=model_path)
    interpreter.allocate_tensors()
    inp = interpreter.get_input_details()[0]
    out = interpreter.get_output_details()[0]
    inp_scale, inp_zero = inp['quantization']
    out_scale, out_zero = out['quantization']
    print(f"Input quantization: scale={inp_scale}, zero_point={inp_zero}")
    print(f"Output quantization: scale={out_scale}, zero_point={out_zero}\n")

    correct = 0
    total = x.shape[0]
    for i in range(total):
        x_float = x[i:i+1]
        # Manually map 0->-128 and 1->127 to match scale=1/255, zp=-128
        x_q = np.where(x_float > 0, 127, -128).astype(np.int8)
        interpreter.set_tensor(inp['index'], x_q)
        interpreter.invoke()
        out_q = interpreter.get_tensor(out['index'])
        out_f = (out_q.astype(np.float32) - out_zero) * out_scale
        if np.argmax(out_f) == np.argmax(y[i]):
            correct += 1
    acc = correct / total
    print(f"Quantized int8 model accuracy: {acc:.4%} ({correct}/{total})\n")

evaluate_tflite_int8("mnist_int8_full.tflite", x_test, y_test)
#
## --- 6) Export quantized weights to Quartus .mif files ---
#out_dir = pathlib.Path("quartus_weights_mif")
#out_dir.mkdir(exist_ok=True)
#print("Exporting weights to Quartus .mif files in", out_dir)
#
#interpreter = tf.lite.Interpreter(model_path="mnist_int8.tflite")
#
##print(interpreter._get_ops_details())
##
### Get number of subgraphs
##num_subgraphs = interpreter._get_ops_details()
##
### Allocate tensors for each subgraph
##for i in range(len(num_subgraphs)):
##    interpreter._interpreter.SetSubgraph(i)
#interpreter.allocate_tensors()
#
## Get all tensor details
#tensor_details = interpreter.get_tensor_details()
#
# # Filter for *constant* weight tensors only — ignore activation/output tensors.
# # Converter‑generated constant weights typically carry the prefix "tfl.pseudo_qconst".
#weight_tensors = [
#	t for t in tensor_details
#	if t['quantization_parameters']['scales'].size > 0
#	and t['name'].startswith("tfl.pseudo_qconst")
#]
#
## ---- Order weight tensors by layer execution sequence --------------
## tf.lite.Interpreter._get_ops_details() yields the ops in the exact
## order the runtime will execute them.  Each op lists its input‑tensor
## indices.  Walk through that list and note the first time each
## weight‑tensor index appears; sort weights by that rank.
#ops_details = interpreter._get_ops_details()  # private but fine for tooling
#exec_order = {}
#for rank, op in enumerate(ops_details):
#    for t_idx in op["inputs"]:
#        if t_idx not in exec_order:
#            exec_order[t_idx] = rank  # first touch = execution order
#
## Give any tensors that never appear in an op input a large rank
## so they fall to the end but are still processed.
#weight_tensors = sorted(
#    weight_tensors,
#    key=lambda t: exec_order.get(t["index"], 1_000_000)
#)
## --------------------------------------------------------------------
#
#txt = out_dir / f"weight_tensors.txt"
#with open(txt, "w") as f:
#    f.write(f"{weight_tensors}\n")
#
#for idx, tensor in enumerate(weight_tensors):
#    # Get the quantized weights
#    weights = interpreter.get_tensor(tensor['index'])
#    scale = tensor['quantization_parameters']['scales'][0]
#    zero_point = tensor['quantization_parameters']['zero_points'][0]
#
#    # Reshape conv kernels if needed (kh,kw,in_ch,out_ch)->(out_ch,in_ch,kh,kw)
#    if len(weights.shape) == 4:
#        weights = np.transpose(weights, (3,2,0,1))
#
#    # Flatten the weights
#    q = weights.flatten()
#    depth = q.size
#    width = 8
#
#    # Encode the original tensor name in the filename for clarity
#    safe_name = tensor["name"].replace("/", "_").replace(";", "_")
#    mif = out_dir / f"{idx:02d}_{safe_name}.mif"
#    with open(mif, "w") as f:
#        f.write(f"DEPTH = {depth};\n")
#        f.write(f"WIDTH = {width};\n")
#        f.write("ADDRESS_RADIX = HEX;\n")
#        f.write("DATA_RADIX = HEX;\n")
#        f.write("CONTENT BEGIN\n")
#        for addr, val in enumerate(q):
#            f.write(f"    {addr:04X} : {(val & 0xFF):02X};\n")
#        f.write("END;\n")
#
#    print(f"Written {mif} (scale={scale:.6f}, zero_point={zero_point})")
#
#
## === EXTRA: Preview first‑convolution feature maps ==========================
## Grab the first test image, quantize it with the same scale/zero‑point as the
## model input, run one inference, and then pull out the tensor that contains
## the output of the *first* convolution (index 8 in this model).
#print("\n--- First‑convolution output (de‑quantized) for test image 0 ---")
#feat_interp = tf.lite.Interpreter(
#    model_path="mnist_int8.tflite",
#    experimental_preserve_all_tensors=True   # <-- key flag
#)
#feat_interp.allocate_tensors()
#
## 2.  Quantise one test image exactly like the model expects
#in_info          = feat_interp.get_input_details()[0]
#in_scale, in_zp  = in_info["quantization"]
#img0_q = np.clip(
#    np.round(x_test[0:1] / in_scale + in_zp),
#    -128, 127
#).astype(np.int8)
#feat_interp.set_tensor(in_info["index"], img0_q)
#
## 3.  Run the network once
#feat_interp.invoke()
#
## 4.  Locate the first‑convolution output tensor
#conv1_details = next(
#    t for t in feat_interp.get_tensor_details()
#    if t["name"].startswith("sequential_1/conv2d_1/Relu")
#)
#conv1_q = feat_interp.get_tensor(conv1_details["index"])
#
## 5.  De‑quantise to float so it’s human‑readable
#c_scale, c_zp = conv1_details["quantization"]
#conv1_i = conv1_q
#conv1_f = (conv1_q.astype(np.float32) - c_zp) * c_scale
#
## 6.  Inspect a small patch (top‑left 5×5 of channel 0)
#print("\n--- First‑convolution output (de‑quantised) for test image 0 ---")
#print(conv1_i[0,:,:,0])
#print(conv1_f[0,:,:,0])
##print(conv1_f[0, :5, :5, 0])