import tensorflow as tf
import numpy as np
import tf_keras as keras

num_classes = 10
input_shape = (28, 28, 1)
(x_train, y_train), (x_test, y_test) = keras.datasets.mnist.load_data()
# binarize to 0/1 and cast to float32
x_train = np.where(x_train > 127, 1, 0).astype(np.int8)
x_test  = np.where(x_test  > 127, 1, 0).astype(np.int8)
# add channel dim
x_train = np.expand_dims(x_train, -1)
x_test  = np.expand_dims(x_test, -1)

y_train = keras.utils.to_categorical(y_train, num_classes)
y_test  = keras.utils.to_categorical(y_test, num_classes)

model = keras.Sequential([
	keras.Input(shape=input_shape),
	keras.layers.Conv2D(12,(3,3),activation="relu"),
	keras.layers.MaxPooling2D((2,2)),
	keras.layers.Flatten(),
	keras.layers.Dense(num_classes,activation="softmax"),
])
model.compile(
    loss="categorical_crossentropy",
    optimizer="adam",
    metrics=["accuracy"]
)

model.fit(x_train, y_train, epochs=15, batch_size=128)

# ensure all layers are built before folding biases
model.build(input_shape=(None, *input_shape))

# --- Fold input zero-point shift into first Conv2D bias for int8 I/O ---
# original int8 zero-point from calibration
old_zp = -128
# get the first Conv2D layer (after the Input layer)
conv_layer = model.layers[1]
W, b = conv_layer.get_weights()
# sum the kernel weights over spatial and input-channel axes for each output filter
kernel_sum = np.sum(W, axis=(0,1,2))  # shape: (num_filters,)
# compute bias correction: shift = -old_zp
delta_b = -old_zp * kernel_sum
# apply folded bias
b_folded = b + delta_b
conv_layer.set_weights([W, b_folded])

import tensorflow_model_optimization as tfmot

# Convert your trained model to a Functional model:
#inputs = tf.keras.Input(shape=(28,28,1))
#outputs = model(inputs)
#func_model = tf.keras.Model(inputs, outputs)

# Apply QAT:
# Annotate the functional model for QAT
annotated_model = tfmot.quantization.keras.quantize_annotate_model(model)
# Apply quantization transforms to create the QAT-ready model
qat_model = tfmot.quantization.keras.quantize_apply(annotated_model)
qat_model.compile(optimizer="adam",
                  loss="categorical_crossentropy",
                  metrics=["accuracy"])
qat_model.fit(x_train, y_train, epochs=2, batch_size=128, validation_split=0.1)
print("QAT accuracy:", qat_model.evaluate(x_test, y_test))

def representative_data_gen():
    for val in tf.data.Dataset.from_tensor_slices(x_train[:100]).batch(1):
        yield [tf.cast(val, tf.float32)]

converter = tf.lite.TFLiteConverter.from_keras_model(qat_model)
converter.optimizations = [tf.lite.Optimize.DEFAULT]
converter.representative_dataset = representative_data_gen
converter.target_spec.supported_ops = [tf.lite.OpsSet.TFLITE_BUILTINS_INT8]

# Force your I/O to uint8 with zero_point=0, scale=1
converter.inference_input_type  = tf.int8
converter.inference_output_type = tf.int8
converter.quantized_input_stats = {
    qat_model.input_names[0]: (0.0, 1.0)
}

tflite_model = converter.convert()

with open("mnist_int8_zp.tflite", "wb") as f:
    f.write(tflite_model)
print("Saved INT8 TFLite model to mnist_int8_zp.tflite\n")

def evaluate_tflite_int8(model_path, x, y):
    interpreter = tf.lite.Interpreter(model_path=model_path)
    interpreter.allocate_tensors()
    inp = interpreter.get_input_details()[0]
    out = interpreter.get_output_details()[0]
    inp_scale, inp_zero = inp['quantization']
    out_scale, out_zero = out['quantization']
    print(f"Input quantization: scale={inp_scale}, zero_point={inp_zero}")
    print(f"Output quantization: scale={out_scale}, zero_point={out_zero}\n")

    correct = 0
    total = x.shape[0]
    for i in range(total):
        x_float = x[i:i+1]
        x_q = np.clip(
            np.round(x_float / inp_scale + inp_zero),
            -128, 127
        ).astype(np.int8)
        interpreter.set_tensor(inp['index'], x_q)
        interpreter.invoke()
        out_q = interpreter.get_tensor(out['index'])
        out_f = (out_q.astype(np.float32) - out_zero) * out_scale
        if np.argmax(out_f) == np.argmax(y[i]):
            correct += 1
    acc = correct / total
    print(f"Quantized int8 model accuracy: {acc:.4%} ({correct}/{total})\n")

evaluate_tflite_int8("mnist_int8_zp.tflite", x_test, y_test)