import numpy as np
import tensorflow as tf
import tensorflow.keras as keras
import pathlib

num_classes = 10
input_shape = (28, 28, 1)
(x_train, y_train), (x_test, y_test) = keras.datasets.mnist.load_data()
# binarize to 0/1 and cast to float32
x_train = np.where(x_train > 127, 127, -128).astype(np.int8)
x_test  = np.where(x_test  > 127, 127, -128).astype(np.int8)
# add channel dim
x_train = np.expand_dims(x_train, -1)
x_test  = np.expand_dims(x_test, -1)

print("\n=== Processing test image through network ===")
feat_interp = tf.lite.Interpreter(
	model_path="mnist_int8.tflite",
	experimental_preserve_all_tensors=True
)
feat_interp.allocate_tensors()

# Print all tensor names for debugging
print("\nAvailable tensors in model:")
for tensor in feat_interp.get_tensor_details():
	print(f"Name: {tensor['name']}, Shape: {tensor['shape']}")

# Get input details and set input tensor
in_info = feat_interp.get_input_details()[0]
in_scale, in_zp = in_info["quantization"]
feat_interp.set_tensor(in_info["index"], x_test[0:1])

input_dir = pathlib.Path("input")
input_dir.mkdir(exist_ok=True)

fname = input_dir / f"x_test"

with open(fname, "w") as f:
	for val in x_test[0:1]:
		for l1 in val:
			for l2 in l1:
				for l3 in l2:
					f.write(f"{l3}\n")

print(f"\nInput image:")
print(f"Shape: {x_test[0:1].shape}")
print(f"Input scale: {in_scale}")
print(f"Input zero-point: {in_zp}")
print(f"True label: {y_test[0]}")

# Run inference (this processes through all layers)
print("\nRunning inference through all layers...")
feat_interp.invoke()

# Get and save convolution layer output
print("\n=== Convolution Layer Output ===")
conv1_details = next(
	t for t in feat_interp.get_tensor_details()
	if t["name"].startswith("sequential_1/conv2d_1/Relu")
)
conv1_q = feat_interp.get_tensor(conv1_details["index"])
c_scale, c_zp = conv1_details["quantization"]
print(f"Conv output shape: {conv1_q.shape}")
print(f"Conv scale: {c_scale}")
print(f"Conv zero-point: {c_zp}")
print(f"Conv details: {conv1_details}")

# Save convolution outputs
conv_dir = pathlib.Path("conv_layer_output")
conv_dir.mkdir(exist_ok=True)

for filter in range(12):
	quantized_filter = conv1_q[0,:,:,filter]
	full_filter = (quantized_filter.astype(np.float32) - c_zp) * c_scale

	fname = conv_dir / f"quantized_{filter}"
	with open(fname, "w") as f:
		for row in quantized_filter:
			for col in row:
				f.write(f"{col},")
			f.write("\n")

	fname = conv_dir / f"full_{filter}"
	with open(fname, "w") as f:
		for row in full_filter:
			for col in row:
				f.write(f"{col},")
			f.write("\n")

# Get and save pooling layer output
print("\n=== Pooling Layer Output ===")
pool_details = next(
	t for t in feat_interp.get_tensor_details()
	if t["name"].startswith("sequential_1/max_pooling2d_1/MaxPool")
)
pool_q = feat_interp.get_tensor(pool_details["index"])
p_scale, p_zp = pool_details["quantization"]
print(f"Pool output shape: {pool_q.shape}")
print(f"Pool scale: {p_scale}")
print(f"Pool zero-point: {p_zp}")
print(f"Pool details: {pool_details}")

pool_dir = pathlib.Path("pool_layer_output")
pool_dir.mkdir(exist_ok=True)

for filter in range(12):
	quantized_filter = pool_q[0,:,:,filter]
	full_filter = (quantized_filter.astype(np.float32) - p_zp) * p_scale

	fname = pool_dir / f"quantized_{filter}"
	with open(fname, "w") as f:
		for row in quantized_filter:
			for col in row:
				f.write(f"{col},")
			f.write("\n")

	fname = pool_dir / f"full_{filter}"
	with open(fname, "w") as f:
		for row in full_filter:
			for col in row:
				f.write(f"{col},")
			f.write("\n")

# Get and save dense layer output
print("\n=== Dense Layer Output ===")
# Try different possible dense layer names
dense_names = [
	"sequential_1/dense_1/Relu",
	"sequential_1/dense_1/BiasAdd",
	"sequential_1/dense_1/MatMul",
	"dense_1/Relu",
	"dense_1/BiasAdd",
	"dense_1/MatMul"
]

dense_details = None
for name in dense_names:
	try:
		dense_details = next(
			t for t in feat_interp.get_tensor_details()
			if t["name"].startswith(name)
		)
		print(f"Found dense layer with name: {dense_details['name']}")
		break
	except StopIteration:
		continue

if dense_details is None:
	print("Could not find dense layer output. Available tensors:")
	for tensor in feat_interp.get_tensor_details():
		print(f"Name: {tensor['name']}, Shape: {tensor['shape']}")
	raise Exception("Dense layer output not found")

dense_q = feat_interp.get_tensor(dense_details["index"])
d_scale, d_zp = dense_details["quantization"]
print(f"Dense output shape: {dense_q.shape}")
print(f"Dense scale: {d_scale}")
print(f"Dense zero-point: {d_zp}")
print(f"Dense details: {dense_details}")

# Save dense outputs
dense_dir = pathlib.Path("dense_layer_output")
dense_dir.mkdir(exist_ok=True)

# Save quantized outputs
fname = dense_dir / "dense_quantized"
with open(fname, "w") as f:
	for val in dense_q[0]:
		f.write(f"{val},")
	f.write("\n")

# Save dequantized outputs
dense_f = (dense_q.astype(np.float32) - d_zp) * d_scale
fname = dense_dir / "dense_dequantized"
with open(fname, "w") as f:
	for val in dense_f[0]:
		f.write(f"{val},")
	f.write("\n")


print("\n=== Final Prediction ===")
prediction = np.argmax(dense_f[0])
print(f"Predicted class: {prediction}")
print(f"True class: {y_test[0]}")
print(f"Correct prediction: {prediction == y_test[0]}")

# === Dump all tensors from the model ===
print("\n=== Dumping all tensors ===")
all_tensors_dir = pathlib.Path("all_tensors")
all_tensors_dir.mkdir(exist_ok=True)
for detail in feat_interp.get_tensor_details():
	idx = detail["index"]
	raw_name = detail["name"]
	safe_name = raw_name.replace("/", "_").replace(":", "_")
	try:
		tensor_data = feat_interp.get_tensor(idx)
	except Exception as e:
		print(f"Skipping tensor {raw_name}: {e}")
		continue
	print(f"Tensor '{raw_name}' (idx={idx}), shape={tensor_data.shape}, dtype={tensor_data.dtype}")
	out_path = all_tensors_dir / f"{safe_name}.txt"
	# Save tensor data as text: flatten multi-dimensional arrays
	flat = tensor_data.flatten()
	# Choose integer or float format based on dtype
	fmt = '%d' if np.issubdtype(tensor_data.dtype, np.integer) else '%.6f'
	np.savetxt(out_path, flat, fmt=fmt)


	if 'qconst3' in raw_name:
		if np.issubdtype(tensor_data.dtype, np.integer):
			# Cast to uint8 so negative int8 values map to two‑s complement 0‑255,
			# then save each element as a zero‑padded two‑digit hex string.
			flat_to_write = flat.astype(np.uint8)
			fmt = '%02X'   # two‑digit hex
		else:
			flat_to_write = flat
			fmt = '%.6f'
		np.savetxt(f"{out_path}-hex", flat, fmt=fmt)