/* * Device driver for the VGA video generator
 *
 * A Platform device implemented using the misc subsystem
 *
 * Stephen A. Edwards
 * Columbia University
 *
 * References:
 * Linux source: Documentation/driver-model/platform.txt
 *               drivers/misc/arm-charlcd.c
 * http://www.linuxforu.com/tag/linux-device-drivers/
 * http://free-electrons.com/docs/
 *
 * "make" to build
 * insmod vga_ball.ko
 *
 * Check code style with
 * checkpatch.pl --file --no-tree vga_ball.c
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/platform_device.h>
#include <linux/miscdevice.h>
#include <linux/slab.h>
#include <linux/io.h>
#include <linux/of.h>
#include <linux/of_address.h>
#include <linux/fs.h>
#include <linux/uaccess.h>
#include "accelerator.h"

#define DRIVER_NAME "accelerator"

/* Device registers */

#define WRITE_INPUT_LAYER1(x) (x)
#define READ_OUTPUT_LAYER1(x) (x + 1)
#define READ_OUTPUT_VALID_LAYER1(x) (x + 2)
#define WRITE_OUTPUT_START_LAYER1(x) (x + 3)
#define WRITE_OUTPUT_LOADDONE_LAYER1(x) (x + 4)
#define WRITE_INPUT_LAYER2(x) (x + 5)
#define READ_OUTPUT_LAYER2(x) (x + 6)
#define READ_OUTPUT_VALID_LAYER2(x) (x + 7)
#define WRITE_OUTPUT_START_LAYER2(x) (x + 8)
#define WRITE_OUTPUT_LOADDONE_LAYER2(x) (x + 9)

/*
 * Information about our device
 */
/*
struct accelerator_dev {
	struct resource res; // Resource: our registers
	void __iomem *virtbase; // Where registers can be accessed in memory
		//u8 layer1OutputBuf[LAYER1_OUTPUT_SIZE];
		//uint8_t layer1Idx;
		short layer1OutputComplete;
		short layer1ValidFlag;
		short layer1StartFlag;
		short layer1LoadDoneFlag;
		u8 readDataLayer1;

		short layer2OutputComplete;
		short layer2ValidFlag;
		short layer2StartFlag;
		short layer2LoadDoneFlag;
		u8 readDataLayer2;

		int running;
		int writeInputData;
} dev;
*/

struct accelerator_dev {
	struct miscdevice misc_dev; /* miscdevice for this instance*/
	struct resource res; /* Resource: our registers */
	void __iomem *virtbase; /* Where registers can be accessed in memory */
	char *name; /* Identifier for different accelerator files */

	short outputComplete;
	short validFlag;
	short startFlag;
	short loadDoneFlag;
	u8 readData;

	int running;
	int writeInputData;
} dev;

static void write_input_layer(int data, int layer)
{
	if (layer == 1) {
		iowrite8(data, WRITE_INPUT_LAYER1(dev.virtbase));
	} else if (layer == 2) {
		iowrite8(data, WRITE_INPUT_LAYER2(dev.virtbase));
	}
	dev.writeInputData = data;
}

static void write_start_flag(int val, int layer)
{
	if (layer == 1) {
		iowrite8(val, WRITE_OUTPUT_START_LAYER1(dev.virtbase));
		dev.layer1StartFlag = val;
	} else if (layer == 2) {
		iowrite8(val, WRITE_OUTPUT_START_LAYER2(dev.virtbase));
		dev.layer2StartFlag = val;
	}
}

static void write_load_complete_flag(int val, int layer)
{
	if (layer == 1) {
		iowrite8(val, WRITE_OUTPUT_LOADDONE_LAYER1(dev.virtbase));
		dev.layer1StartFlag = val;
	} else if (layer == 2) {
		iowrite8(val, WRITE_OUTPUT_LOADDONE_LAYER2(dev.virtbase));
		dev.layer2StartFlag = val;
	}
}


static void read_valid_layer(int layer)
{
	if (layer == 1)
		dev.layer1ValidFlag = ioread8(READ_OUTPUT_VALID_LAYER1(dev.virtbase));
	else if (layer == 2)
		dev.layer2ValidFlag = ioread8(READ_OUTPUT_VALID_LAYER2(dev.virtbase));
}

static void read_output_layer(int layer)
{
	if (layer == 1)
		dev.readDataLayer1 = ioread8(READ_OUTPUT_LAYER1(dev.virtbase));
	else if (layer == 2)
		dev.readDataLayer2 = ioread8(READ_OUTPUT_LAYER2(dev.virtbase));
}

/*
 * Handle ioctl() calls from userspace:
 * Read or write the segments on single digits.
 * Note extensive error checking of arguments
 */
static long accelerator_ioctl(struct file *f, unsigned int cmd, unsigned long arg)
{
	struct accelerator_dev *idev = f->private_data;	
	accelerator_arg_t vla;

	if (!idev) {
		pr_err(DRIVE_NAME ": ioctl called with no instance data\n");
		return -ENODEV;
	}

	switch (cmd) {
	case ACCEL_WRITE_INPUT:
		if (copy_from_user(&vla, (accelerator_arg_t *) arg,
				   sizeof(accelerator_arg_t)))
			return -EACCES;
		// write_input_layer(&vla.layer1.input_byte, 1);
		iowrite8(vla.input_byte, idev->virtbase + ACCEL_REG_INPUT);
		idev->writeInputData = vla.input_byte;
		break;
	case ACCEL_READ_LAYER1_OUTPUT:
		read_output_layer(1);
		vla.layer1.output_byte = dev.readDataLayer1;
		if (copy_to_user((accelerator_arg_t *) arg, &vla,
				sizeof(accelerator_arg_t)))
			return -EACCES;
		break;
	case ACCEL_READ_LAYER1_VALID_F:
		read_valid_layer(1);
		vla.layer1.valid_flag = dev.layer1StartFlag;
		if (copy_to_user((accelerator_arg_t *) arg, &vla,
				sizeof(accelerator_arg_t)))
			return -EACCES;
		break;
	case ACCEL_WRITE_LAYER1_START_F:
		if (copy_from_user(&vla, (accelerator_arg_t *) arg,
				   sizeof(accelerator_arg_t)))
			return -EACCES;
		write_input_layer(&vla.layer1.start_flag, 1);
		break;
	case ACCEL_WRITE_LAYER1_LOADDONE_F:
		if (copy_from_user(&vla, (accelerator_arg_t *) arg,
				   sizeof(accelerator_arg_t)))
			return -EACCES;
		write_input_layer(&vla.layer1.load_complete_flag, 1);
		break;

	case ACCEL_WRITE_LAYER2_INPUT:
		if (copy_from_user(&vla, (accelerator_arg_t *) arg,
				   sizeof(accelerator_arg_t)))
			return -EACCES;
		write_input_layer(&vla.layer2.input_byte, 2);
		break;
	case ACCEL_READ_LAYER2_OUTPUT:
		read_output_layer(2);
		vla.layer2.output_byte = dev.readDataLayer2;
		if (copy_to_user((accelerator_arg_t *) arg, &vla,
				sizeof(accelerator_arg_t)))
			return -EACCES;
		break;
	case ACCEL_READ_LAYER2_VALID_F:
		read_valid_layer(2);
		vla.layer2.valid_flag = dev.layer2StartFlag;
		if (copy_to_user((accelerator_arg_t *) arg, &vla,
				sizeof(accelerator_arg_t)))
			return -EACCES;
		break;
	case ACCEL_WRITE_LAYER2_START_F:
		if (copy_from_user(&vla, (accelerator_arg_t *) arg,
				   sizeof(accelerator_arg_t)))
			return -EACCES;
		write_input_layer(&vla.layer2.start_flag, 2);
		break;
	case ACCEL_WRITE_LAYER2_LOADDONE_F:
		if (copy_from_user(&vla, (accelerator_arg_t *) arg,
				   sizeof(accelerator_arg_t)))
			return -EACCES;
		write_input_layer(&vla.layer2.load_complete_flag, 2);
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

/* The operations our device knows how to do */
static const struct file_operations accelerator_fops = {
	.owner		    = THIS_MODULE,
	.open           = accelerator_open,
	.close          = accelerator_close,
	.unlocked_ioctl = accelerator_ioctl,
};

/* Information about our device for the "misc" framework -- like a char dev */
/*
static struct miscdevice accelerator_misc_device = {
	.minor		= MISC_DYNAMIC_MINOR,
	.name		= DRIVER_NAME,
	.fops		= &accelerator_fops,
};
*/

/*
 * Initialization code: get resources (registers) and display
 * a welcome message
 */
static int __init accelerator_probe(struct platform_device *pdev)
{
	const struct of_device_id *match;
	const char *device_name_from_dt;
	enum accel_hw_type hw_type; /* hw type identifier */
        struct accelerator_dev *idev; /* pointer to the instance-specific data*/
	int ret;

	match = of_match_device(accelerator_of_match, &pdev->dev);
	hw_type = (enum accel_hw_type)match->data; /* get the type */

	/* Allocate memory for device instance */
	idev = devm_kzalloc(&pdev->dev, sizeof(*idev), GFP_KERNEL);
	if (!idev) {
        	dev_err(&pdev->dev, "Failed to allocate memory for instance\n");
        	return -ENOMEM;
	}
	idev->pdev = pdev; // Store platform_device

	/* Get the address of our registers from the device tree */
	ret = of_address_to_resource(pdev->dev.of_node, 0, &idev->res);
	if (ret) {
		return -ENOENT;
	}

	/* Make sure we can use these registers */
	if (devm_request_mem_region(&pdev->dev, idev->res.start, resource_size(&idev->res),
			       dev_name(&pdev->dev)) == NULL) {
		return -EBUSY;
	}

	/* Arrange access to our registers */
	idev->virtbase = devm_ioremap_nocache(&pdev->dev, idev->res.start, resource_size(&idev->res));
	if (idev->virtbase == NULL) {
		return -ENOMEM;
	}

	/* Identify and probe corresponding hw type
	switch (hw_type) {
		case CONVOLVER:
		dev_info(&pdev->dev, "Probing a Convolver accelerator\n");
		// Specific setup for convolver, e.g., different register offsets
		// iowrite32(INIT_VAL_CONV, idev->virtbase + CONV_SPECIFIC_REG);
		break;
	case POOLING_UNIT:
		dev_info(&pdev->dev, "Probing a Pooling accelerator\n");
		// Specific setup for pooling unit
		// iowrite32(INIT_VAL_POOL, idev->virtbase + POOL_SPECIFIC_REG);
		break;
	default:
		dev_warn(&pdev->dev, "Probing an unknown accelerator type\n");
		// Generic setup or error
		break;
	} */


	/* Configure the miscdevice structure for current instance (as opposed to globally) */
	idev->misc_dev.minor = MISC_DYNAMIC_MINOR;
	idev->misc_dev.fops = &accelerator_fops;

	if (of_property_read_string(pdev->dev.of_node, "label", &device_name_from_dt) == 0) {
		idev->misc_dev.name = devm_kasprintf(&pdev->dev, GFP_KERNEL, "%s", device_name_from_dt);
	} else {
		// Fallback: use platform device name, ensure it's unique enough or add an id
		idev->misc_dev.name = devm_kasprintf(&pdev->dev, GFP_KERNEL, "%s_%d", pdev->name, pdev->id);
	}

	/* Register current instance */
	ret = misc_register(&idev->misc_dev);
	if (ret) {
		return ret;
	}

	/* Store instance-specific data with platform-device */
	platform_set_drvdata(pdev, idev);

	return 0;
/*
out_release_mem_region:
	release_mem_region(dev.res.start, resource_size(&dev.res));
out_deregister:
	misc_deregister(&accelerator_misc_device);
	return ret;
*/
}

/* Find instance specific driver (depending on layer) */
static int accelerator_open(struct inode *inode, struct file *file)
{
	struct miscdevice *miscdev = file->private_data; // get misc device struct
	struct accelerator_dev *idev;

	/* Get the pointer to containing struct accelerator_dev */
	idev = container_of(miscdev, struct accelerator_dev, misc_dev);
	if (!idev) {
		pr_err(DRIVE_NAME ": could not retrieve instance data in open\n");
		return -ENODEV;
	}

	/* Make file data easy to access in other helper functions */
	file->private_data = idev;

	/* Access idev->hw_type for type-specific open logic */
    // dev_info(&idev->pdev->dev, "Device %s opened (type: %d)\n", idev->misc_dev.name, idev->hw_type);

	return 0;
}

/* Clean-up code: release resources for single open instance of the device file */
static int accelerator_close(struct inode *inode, struct file *file)
{
	struct accelerator_dev *idev = file->private_data;
	if (!idev) {
		pr_err(DRIVE_NAME ": could not retrieve instance data in close\n");
		return -ENODEV;
	}

	return 0;
}

/* Clean-up code: release resources for overall driver */
static int accelerator_remove(struct platform_device *pdev)
{
	struct accelerator_dev *idev = platform_get_drvdata(pdev);
	if (idev) misc_deregister(&idev->misc_dev);
	return 0;
}

/* Type handling for device tree and probing */
enum accel_hw_type {
    CONVOLVER,
    DENSE,
};

/* Which "compatible" string(s) to search for in the Device Tree */
#ifdef CONFIG_OF
static const struct of_device_id accelerator_of_match[] = {
	{ .compatible = "csee4840,convolver-1.0", .data = (void *)CONVOLVER },
	{ .compatible = "csee4840,dense-1.0", .data = (void *)DENSE },
	{},
};
MODULE_DEVICE_TABLE(of, accelerator_of_match);
#endif

/* Information for registering ourselves as a "platform" driver */
static struct platform_driver accelerator_driver = {
	.driver	= {
		.name	= DRIVER_NAME,
		.owner	= THIS_MODULE,
		.of_match_table = of_match_ptr(accelerator_of_match),
	},
	.probe = accelerator_probe,
	.remove	= accelerator_remove,
};

/* Called when the module is loaded: set things up */
static int __init accelerator_init(void)
{
	pr_info(DRIVER_NAME ": init\n");
	return platform_driver_register(&accelerator_driver);
}

/* Calball when the module is unloaded: release resources */
static void __exit accelerator_exit(void)
{
	platform_driver_unregister(&accelerator_driver);
	pr_info(DRIVER_NAME ": exit\n");
}

module_init(accelerator_init);
module_exit(accelerator_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Stephen A. Edwards, Columbia University");
MODULE_DESCRIPTION("Accelerator");
