/* * Device driver for the VGA video generator
 *
 * A Platform device implemented using the misc subsystem
 *
 * Stephen A. Edwards
 * Columbia University
 *
 * References:
 * Linux source: Documentation/driver-model/platform.txt
 *               drivers/misc/arm-charlcd.c
 * http://www.linuxforu.com/tag/linux-device-drivers/
 * http://free-electrons.com/docs/
 *
 * "make" to build
 * insmod vga_ball.ko
 *
 * Check code style with
 * checkpatch.pl --file --no-tree vga_ball.c
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/platform_device.h>
#include <linux/miscdevice.h>
#include <linux/slab.h>
#include <linux/io.h>
#include <linux/of.h>
#include <linux/of_address.h>
#include <linux/fs.h>
#include <linux/uaccess.h>
#include "control.h"

#define DRIVER_NAME "control"

/* Device registers */
#define IMAGE(x) (x)
#define PROB_0(x) ((x)+4)
#define PROB_1(x) ((x)+8)
#define PROB_2(x) ((x)+12)
#define PROB_3(x) ((x)+16)
#define PROB_4(x) ((x)+20)
#define PROB_5(x) ((x)+24)
#define PROB_6(x) ((x)+28)
#define PROB_7(x) ((x)+32)
#define PROB_8(x) ((x)+36)
#define PROB_9(x) ((x)+40)
#define DENSE_DONE_F(x) ((x)+60)


/*
* Information about our device
*/
struct control_dev {
    struct resource res; /* Resource: our registers */
    void __iomem *virtbase; /* Where registers can be accessed in memory */ 
        uint32_t dense_data [10];
        uint32_t image_data;
        uint32_t dense_done_f;
} dev;

/*
* Write segments of a single digit
* Assumes digit is in range and the device information has been set up
*/
static void write_image_data(uint32_t *image_data)
{
    iowrite32(*image_data, IMAGE(dev.virtbase));
    dev.image_data = *image_data;
}

static void read_dense_done(void)
{
    dev.dense_done_f = ioread32(DENSE_DONE_F(dev.virtbase));
}

static void trigger_dense_addr(void)
{
    for (int i = 0; i < 10; i++) {
        ioread32(PROB_0(dev.virtbase) + i * 4);
    }
}

static void read_dense_data(void)
{
    for (int i = 0; i < 10; i++) {
        dev.dense_data[i] = ioread32(PROB_0(dev.virtbase) + i * 4);
    }
}

/*
* Handle ioctl() calls from userspace:
* Read or write the segments on single digits.
* Note extensive error checking of arguments
*/
static long control_ioctl(struct file *f, unsigned int cmd, unsigned long arg)
{
    control_arg_t vla;

    switch (cmd) {
    case CONTROL_WRITE_IMAGE:
        if (copy_from_user(&vla, (control_arg_t *) arg,
                sizeof(control_arg_t)))
            return -EACCES;
        write_image_data(&vla.image_data);
        break;

    case CONTROL_READ_DENSE_DONE:
        read_dense_done();
        vla.dense_done_f = dev.dense_done_f;
        if (copy_to_user((control_arg_t *) arg, &vla,
                sizeof(control_arg_t)))
            return -EACCES;
        break;

    case CONTROL_TRIGGER_DENSE_ADDR:
        trigger_dense_addr();
        break;

    case CONTROL_READ_DENSE_DATA:
        read_dense_data();
        memcpy(vla.dense_data, dev.dense_data, sizeof(dev.dense_data));
        if (copy_to_user((control_arg_t *) arg, &vla,
                sizeof(control_arg_t)))
            return -EACCES;
        break;

    default:
        return -EINVAL;
    }

    return 0;
}

/* The operations our device knows how to do */
static const struct file_operations control_fops = {
    .owner		= THIS_MODULE,
    .unlocked_ioctl = control_ioctl,
};

/* Information about our device for the "misc" framework -- like a char dev */
static struct miscdevice control_misc_device = {
    .minor		= MISC_DYNAMIC_MINOR,
    .name		= DRIVER_NAME,
    .fops		= &control_fops,
};

/*
* Initialization code: get resources (registers) and display
* a welcome message
*/
static int __init control_probe(struct platform_device *pdev)
{
    //vga_ball_color_t beige = { 0xf9, 0xe4, 0xb7 };
    int ret;

    /* Register ourselves as a misc device: creates /dev/control */
    ret = misc_register(&control_misc_device);

    /* Get the address of our registers from the device tree */
    ret = of_address_to_resource(pdev->dev.of_node, 0, &dev.res);
    if (ret) {
        ret = -ENOENT;
        goto out_deregister;
    }

    /* Make sure we can use these registers */
    if (request_mem_region(dev.res.start, resource_size(&dev.res),
                DRIVER_NAME) == NULL) {
        ret = -EBUSY;
        goto out_deregister;
    }

    /* Arrange access to our registers */
    dev.virtbase = of_iomap(pdev->dev.of_node, 0);
    if (dev.virtbase == NULL) {
        ret = -ENOMEM;
        goto out_release_mem_region;
    }

    /* Set an initial color */
        //write_background(&beige);

    return 0;

out_release_mem_region:
    release_mem_region(dev.res.start, resource_size(&dev.res));
out_deregister:
    misc_deregister(&control_misc_device);
    return ret;
}

/* Clean-up code: release resources */
static int control_remove(struct platform_device *pdev)
{
    iounmap(dev.virtbase);
    release_mem_region(dev.res.start, resource_size(&dev.res));
    misc_deregister(&control_misc_device);
    return 0;
}

/* Which "compatible" string(s) to search for in the Device Tree */
#ifdef CONFIG_OF
static const struct of_device_id control_of_match[] = {
    { .compatible = "csee4840,vga_ball-1.0" },
    {},
};
MODULE_DEVICE_TABLE(of, control_of_match);
#endif

/* Information for registering ourselves as a "platform" driver */
static struct platform_driver control_driver = {
    .driver	= {
        .name	= DRIVER_NAME,
        .owner	= THIS_MODULE,
        .of_match_table = of_match_ptr(control_of_match),
    },
    .remove	= __exit_p(control_remove),
};

/* Called when the module is loaded: set things up */
static int __init control_init(void)
{
    pr_info(DRIVER_NAME ": init\n");
    return platform_driver_probe(&control_driver, control_probe);
}

/* Calball when the module is unloaded: release resources */
static void __exit control_exit(void)
{
    platform_driver_unregister(&control_driver);
    pr_info(DRIVER_NAME ": exit\n");
}

module_init(control_init);
module_exit(control_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Stephen A. Edwards, Columbia University");
MODULE_DESCRIPTION("VGA ball driver");
