#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include <stdatomic.h>
#include <stdbool.h>
#include <errno.h>

#include "control.h" //header file

#include <stdint.h>
#include <stdlib.h>
#include <time.h>

#define IMAGE_SIZE 784

int control_fd;

int load_image_data(uint32_t *image_data_word) {
	control_arg_t vla;
	vla.image_data = *image_data_word;
	if (ioctl(control_fd, CONTROL_WRITE_IMAGE, &vla)) {
		perror("ioctl(CONTROL_WRITE_IMAGE) failed");
		return -1;
	}
	return 0;
}

int read_dense_done_sig(uint32_t *dense_done_status) {
	control_arg_t vla;
	if (ioctl(control_fd, CONTROL_READ_DENSE_DONE, &vla)) {
		perror("ioctl(CONTROL_READ_DENSE_DONE) failed");
		return -1;
	}
	// SEND DATA TO USERSPACE MEMORY FOR IT TO POLL AND CHECK VALUE, backwards
	*dense_done_status = vla.dense_done_f;
	return 0;
}

int trigger_dense_addr(void) {
  control_arg_t vla;
	if (ioctl(control_fd, CONTROL_TRIGGER_DENSE_ADDR, &vla)) {
		perror("ioctl(CONTROL_TRIGGER_DENSE_ADDR) failed");
		return -1;
	}
	return 0;
}

int read_dense_outputs(uint32_t *dense_outputs) {
	control_arg_t vla;
	if (ioctl(control_fd, CONTROL_READ_DENSE_DATA, &vla)) {
		perror("ioctl(CONTROL_READ_DENSE_DATA) failed");
		return -1;
	}
	memcpy(dense_outputs, vla.dense_data, sizeof(vla.dense_data));
	return 0;
}

int main() {
  control_arg_t vla;
  static const char filename[] = "/dev/control";
  int32_t dense_outputs[10];

  struct timespec start_ts, end_ts;
  double total_ms;

  printf("Control Userspace program started\n");

  // Get start time
  if (clock_gettime(CLOCK_MONOTONIC, &start_ts) == -1) {
    perror("clock_gettime start failed");
    return EXIT_FAILURE;
  }

  if ( (control_fd = open(filename, O_RDWR)) == -1 ) {
    fprintf(stderr, "could not open %s\n", filename);
    return -1;
  }

  /* Open image file */
  FILE *im_fp;
  char *image_path = "../model/quartus_weights_mif/x_test.mem";
  im_fp = fopen(image_path, "r");
  if (im_fp == NULL) {
    fprintf(stderr, "could not open %s\n", image_path);
    return -1;
  }
  
  /* Initialize line buffer */
  char line_buff [8];
  int pixel_count = 0;

  /* run fgets(), build image_word and send with ioctl */
  while (fgets(line_buff, sizeof(line_buff), im_fp) && pixel_count < IMAGE_SIZE) {
    /* Create image word to be sent */
    uint32_t image_word = 0;
    
    /* Get building blocks of image word */
    uint8_t flag = (pixel_count == IMAGE_SIZE - 1) ? 1 : 0;
    uint16_t addr = pixel_count;
    uint8_t pixel_data = (line_buff[0] == '1') ? 1 : 0;

    /* Build image word */
    image_word |= (flag & 0x1) << 11; // make sure flag is 1 bit, shift it left 11 bits, and then 'or' it with image_word
    image_word |= (addr & 0x3ff) << 1;
    image_word |= (pixel_data & 0x1);

    /* Send image with ioctl */
    if (load_image_data(&image_word) != 0) {
      perror("ioctl(CONTROL_WRITE_IMAGE) failed");
      return EXIT_FAILURE;
    }

    pixel_count++;
  }

  fclose(im_fp);

  // POLL FOR DENSE DONE SIGNAL
  uint32_t dense_done_status = 0; //at program start, initialize and set low

  while (1) {
    // Error checking for ioctl call
    if (read_dense_done_sig(&dense_done_status) != 0) {
      perror("ioctl(CONTROL_READ_DENSE_DONE) failed");
      return EXIT_FAILURE;
    }

    // when sig high, break out of loop
    if (dense_done_status == 1) {
      break;
    }
    usleep(10000); // wait 10ms
  }

  // SEND ADDRESS TRIGGER_DENSE_ADDRESSES
  if (trigger_dense_addr() != 0) {
    perror("ioctl(CONTROL_TRIGGER_DENSE_ADDR) failed");
    return EXIT_FAILURE;
  }

  // READ OUT DENSE OUTPUTS
  if (read_dense_outputs(dense_outputs) != 0) {
    perror("ioctl(CONTROL_READ_DENSE_DATA) failed");
    return EXIT_FAILURE;
  } else {
    for (int i = 0; i < 10; i++) {
      printf("Dense output %d: %d\n", i, dense_outputs[i]);
    }
  }

  // Get end time
  if (clock_gettime(CLOCK_MONOTONIC, &end_ts) == -1) {
    perror("clock_gettime end failed");
    close(control_fd);
    return EXIT_FAILURE;
  }

  // Calculate the difference in milliseconds
  total_ms = (end_ts.tv_sec - start_ts.tv_sec) * 1000.0;       // Convert secs to ms
  total_ms += (end_ts.tv_nsec - start_ts.tv_nsec) / 1000000.0; // Convert ns to ms and add
  printf("Runtime: %.3f ms\n", total_ms);
  
  if (close(control_fd) == -1) {
    perror("Failed to close /dev/control");
    // Potentially return EXIT_FAILURE or log, but program is ending anyway
  }

  // CLOSE DEVICE KERNEL DRIVER || PROGRAM EXIT (CLOSE(control_fd))?
  printf("Control Userspace program terminating\n");
  return EXIT_SUCCESS;
}
